/***************************************
  $Header: /home/amb/wwwoffle/RCS/wwwoffles.c 2.26 1997/12/30 16:19:13 amb Exp $

  WWWOFFLE - World Wide Web Offline Explorer - Version 2.0a.
  A server to fetch the required pages.
  ******************/ /******************
  Written by Andrew M. Bishop

  This file Copyright 1996,97 Andrew M. Bishop
  It may be distributed under the GNU Public License, version 2, or
  any higher version.  See section COPYING of the GNU Public license
  for conditions under which this file may be redistributed.
  ***************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>

#include "wwwoffle.h"
#include "misc.h"
#include "proto.h"
#include "config.h"
#include "sockets.h"
#include "errors.h"


static void uninstall_sighandlers(void);


/*+ The mode of operation of the server. +*/
typedef enum _Mode
{
 None,                          /*+ Undecided. +*/
 Real,                          /*+ From server host to cache and client. +*/
 RealNoCache,                   /*+ From server host to client. +*/
 RealRefresh,                   /*+ Refresh the page, forced from index. +*/
 Fetch,                         /*+ From server host to cache. +*/
 Spool,                         /*+ From cache to client. +*/
 SpoolGet,                      /*+ Not in cache so record request in outgoing. +*/
 SpoolWillGet,                  /*+ Not in cache but is in outgoing. +*/
 SpoolRefresh,                  /*+ Refresh the page, forced from index. +*/
 SpoolPragma                    /*+ Refresh the page, forced from browser by 'Pragma: no-cache'. +*/
}
Mode;


/*++++++++++++++++++++++++++++++++++++++
  The main server program.

  int wwwoffles Returns the exit status.

  int online Whether the demon is online or not.

  int browser Set to true if there is a browser.

  int client The file descriptor of the client.
  ++++++++++++++++++++++++++++++++++++++*/

int wwwoffles(int online,int browser,int client)
{
 int outgoing=-1,spool=-1;

 char *request=NULL,*reply=NULL,*url=NULL;
 int replylen=0;

 char *mirrorProto=NULL,*mirrorHost=NULL;

 URL *Url;

 Mode mode=None;
 int fetch_again=0;

 /* Initialise things. */

 uninstall_sighandlers();

 InitErrorHandler("wwwoffles",-1,-1); /* change name nothing else */

 if(online && browser)
    mode=Real;
 else if(online && !browser)
    mode=Fetch;
 else if(!online && browser)
    mode=Spool;
 else
    PrintMessage(Fatal,"Started in a mode that is not allowed.");

 /* Set up the client file. */

 if(client<0 && mode!=Fetch)
   {PrintMessage(Warning,"Cannot use client file descriptor %d.",client);exit(1);}

 /* Set up the outgoing file. */

 if(mode==Fetch)
   {
    outgoing=OpenOutgoingSpoolFile(1);
    init_buffer(outgoing);

    if(outgoing==-1)
      {PrintMessage(Inform,"No more outgoing requests.");exit(3);}
   }

 /* Get the URL from the request. */

 if(mode==Real || mode==Spool)
    request=ParseRequest(client,&url);
 else /* mode==Fetch */
    request=ParseRequest(outgoing,&url);

 if(!url)
   {
    if(mode!=Fetch)
       ServerError(client,request?"Cannot parse the HTTP request":"The HTTP request was empty");
    PrintMessage(Warning,"Could not parse HTTP request (%s).",request?"Parse error":"Empty request");
    exit(1);
   }

 PrintMessage(Inform,"URL='%s'.",url);

 Url=SplitURL(url);

 PrintMessage(Debug,"proto='%s'; host='%s'; path='%s'; args='%s'.",Url->proto,Url->host,Url->path,Url->args);

 if(!Url->Protocol)
   {
    if(mode==Real || mode==Spool)
       IllegalProto(client,Url->name,Url->proto);
    PrintMessage(Inform,"The protocol '%s' is not available.",Url->proto);
    exit(1);
   }

 /* Change the mode based on the URL as required. */
 /* (At this point mode is Spool, Real or Fetch only.) */

 if(IsNotGot(Url->proto,Url->host,Url->path))
   {
    char* colon=strchr(Url->host,':');
    if(colon) *colon=0;
    if(mode==Real || mode==Spool)
       RemoteHostError(client,Url->name,"is on the list of hosts and paths not to get.");
    PrintMessage(Inform,"The server '%s://%s' and/or path '%s' is on the list not to get.",Url->proto,Url->host,Url->path);
    exit(0);
   }

 if(Url->local)
   {
    int newurl=0;

    if(!strncmp("/refresh",Url->path,8))
      {
       url=RefreshPage(client,Url->path,Url->args,Url->name,request,online);

       if(url==(char*)1)
          newurl=-1;
       else if(url)
         {
          FreeURL(Url);
          Url=SplitURL(url);

          if(Url->args && *Url->args=='!')
            {
             PrintMessage(Inform,"It is not possible to refresh a URL that was posted.");
             if(mode!=Fetch)
                CantRefreshPosted(client,Url->name);
             exit(0);
            }
          else if(!*Url->host || IsLocalNetHost(Url->host))
             mode=RealNoCache;
          else if(mode==Real)
            {
             mode=RealRefresh;
             DeleteWebpageSpoolFile(Url,0);
            }
          else if(mode==Spool)
            {
             if(ExistsWebpageSpoolFile(Url))
                mode=SpoolRefresh;
             else
                if(ExistsOutgoingSpoolFile(Url))
                   mode=SpoolWillGet;
                else
                   mode=SpoolGet;
            }
          newurl=1;
         }
      }
    else if(mode==Fetch)
       PrintMessage(Inform,"The request to fetch a page from the local host is ignored.");
    else if(!strcmp(Url->path,"/") && !Url->args)
       WelcomePage(client);
    else if(!strncmp("/index/",Url->path,7))
       IndexPage(client,&Url->path[7],Url->args);
    else if(!strncmp("/control/",Url->path,9))
       ControlPage(client,&Url->path[9],Url->args,request);
    else
      {
       int i;

       for(i=0;i<NProtocols;i++)
          if(!strncmp(Protocols[i].name,Url->pathp+1,strlen(Protocols[i].name)) &&
             Url->pathp[strlen(Protocols[i].name)+1]=='/')
            {
             url=(char*)malloc(strlen(Url->pathp)+4);
             Url->pathp[strlen(Protocols[i].name)+1]=0;
             sprintf(url,"%s://%s",Url->pathp+1,&Url->pathp[strlen(Protocols[i].name)+2]);
             Url->pathp[strlen(Protocols[i].name)+1]='/';
             newurl=1;
             break;
            }

       if(!newurl)
          IllegalPage(client,Url->pathp);

       FreeURL(Url);
       Url=SplitURL(url);
      }

    if(!newurl)
       exit(0);
    else if(newurl==-1)
       exit(4);

    PrintMessage(Inform,"new URL='%s'.",url);
    PrintMessage(Debug,"new proto='%s'; host='%s'; path='%s'; args='%s'.",Url->proto,Url->host,Url->path,Url->args);

    if(!Url->Protocol)
      {
       if(mode==Real || mode==Spool)
          IllegalProto(client,Url->name,Url->proto);
       PrintMessage(Inform,"The protocol '%s' is not available.",Url->proto);
       exit(1);
      }
   }

 if(IsMirrored(Url->proto,Url->host,&mirrorProto,&mirrorHost))
   {
    char *newurl=(char*)malloc(strlen(mirrorProto)+strlen(mirrorHost)+strlen(Url->pathp)+8);

    sprintf(newurl,"%s://%s%s",mirrorProto,mirrorHost,Url->pathp);

    url=newurl;

    FreeURL(Url);
    Url=SplitURL(url);

    PrintMessage(Inform,"mirrored URL='%s'.",url);
    PrintMessage(Debug,"mirrored proto='%s'; host='%s'; path='%s'; args='%s'.",Url->proto,Url->host,Url->path,Url->args);

    if(!Url->Protocol)
      {
       if(mode==Real || mode==Spool)
          IllegalProto(client,Url->name,Url->proto);
       PrintMessage(Inform,"The protocol '%s' is not available.",Url->proto);
       exit(1);
      }
   }

 if(IsLocalNetHost(Url->host))
   {
    if(mode==Real || mode==Spool)
       mode=RealNoCache;
    else
      {
       PrintMessage(Inform,"The request to fetch a page from the local network host '%s' is ignored.",Url->host);
       exit(0);
      }
   }
 else if(mode==Real && IsNotCached(Url->proto,Url->host,Url->path))
   {
    mode=RealNoCache;
   }

 if(!strncasecmp(request,"POST",4))
   {
    if(mode==Spool)
       mode=SpoolGet;
   }
 else if(Url->args && *Url->args=='!')
   {
    if(mode==Real)
       mode=Spool;
    else if(mode==Fetch)
      {
       PrintMessage(Inform,"It is not possible to fetch a URL that was posted.");
       exit(0);
      }
   }

 if(strstr(request,"\nIf-Modified-Since:"))
   {
    char *bol=strstr(request,"\nIf-Modified-Since:")+1,*eol=strchr(bol,'\n');
    *eol++=0;

    if(mode==Spool)
      {
       spool=OpenWebpageSpoolFile(1,Url);

       if(spool!=-1)
         {
          struct stat buf;
          long since;

          buf.st_mtime=time(NULL)+1;
          fstat(spool,&buf);
          since=DateToTimeT(bol+19);

          close(spool);

          if(since>buf.st_mtime)
            {NotModified(client);exit(0);}
         }
      }

    while(*eol)
       *bol++=*eol++;
    *bol=0;
   }

 if(PragmaNoCache && strstr(request,"\nPragma: no-cache"))
   {
    if(mode==Spool || mode==SpoolGet)
      {
       if(ExistsWebpageSpoolFile(Url))
          mode=SpoolPragma;
       else
          if(ExistsOutgoingSpoolFile(Url))
             mode=SpoolWillGet;
          else
             mode=SpoolGet;
      }
    /* (mode==Fetch || mode==Real) are left unchanged, not modified as below. */
   }
 else if(mode==Spool && !ExistsWebpageSpoolFile(Url))
   {
    if(ExistsOutgoingSpoolFile(Url))
       mode=SpoolWillGet;
    else
       mode=SpoolGet;
   }
 else if(mode==Fetch && ExistsWebpageSpoolFile(Url))
   {
    spool=OpenWebpageSpoolFile(1,Url);

    if(RequestChanges(spool,&request)==1)
       TouchWebpageSpoolFile(Url);
    else
       exit(0);

    close(spool);
   }
 else if(mode==Real && ExistsWebpageSpoolFile(Url))
   {
    spool=OpenWebpageSpoolFile(1,Url);

    if(RequestChanged && RequestChanges(spool,&request)==1)
       TouchWebpageSpoolFile(Url);
    else
       mode=Spool;

    close(spool);
   }

 /* Set up the file descriptor for the spool file. */

 if(mode==Real || mode==Fetch)
   {
    spool=OpenWebpageSpoolFile(0,Url);

    if(spool==-1)
      {
       if(mode==Real)
          ServerError(client,"Cannot open the spooled web page to write.");
       PrintMessage(Warning,"Cannot open the spooled web page to write.");
       exit(1);
      }
   }
 else if(mode==Spool || mode==SpoolPragma)
   {
    spool=OpenWebpageSpoolFile(1,Url);
    init_buffer(spool);

    if(spool==-1)
      {
       ServerError(client,"Cannot open the spooled web page to read.");
       PrintMessage(Warning,"Cannot open the spooled web page to read.");
       exit(1);
      }
   }

 /* Set up the outgoing file. */

 if(mode==SpoolGet || mode==SpoolRefresh || mode==SpoolPragma)
   {
    outgoing=OpenOutgoingSpoolFile(0);

    if(outgoing==-1)
      {
       ServerError(client,"Cannot open the outgoing request to write.");
       PrintMessage(Warning,"Cannot open the outgoing request to write.");
       exit(1);
      }
   }

 /* Open the connection to the server host. */

 if(mode==Real || mode==RealNoCache || mode==Fetch)
   {
    char *err=(Url->Protocol->open)(Url);

    if(err)
      {
       if(mode==Fetch || mode==Real)
         {
          lseek(spool,0,SEEK_SET);
          ftruncate(spool,0);
          RemoteHostError(spool,Url->name,err);
         }
       if(mode==Real || mode==RealNoCache)
          RemoteHostError(client,Url->name,err);
       exit(1);
      }
   }

 /* Modify the header (Censor / Cannonicalise URL / POST / HTTP-1.1 etc). */

 if(mode==Real || mode==RealNoCache || mode==SpoolGet || mode==SpoolRefresh || mode==SpoolPragma)
    request=ModifyRequest(Url,request);

 /* Write request to remote server or outgoing file. */

 if(mode==Real || mode==RealNoCache || mode==Fetch)
   {
    char *err=(Url->Protocol->request)(Url,request);

    if(err)
      {
       if(mode==Real || mode==Fetch)
         {
          lseek(spool,0,SEEK_SET);
          ftruncate(spool,0);
          RemoteHostError(spool,Url->name,err);
         }
       if(mode==Real || mode==RealNoCache)
          RemoteHostError(client,Url->name,err);
       exit(1);
      }
    else if(mode==Fetch && client!=-1)
       write_formatted(client,"Fetching %s\n",Url->name);
   }
 else if(mode==SpoolGet || mode==SpoolRefresh || mode==SpoolPragma)
   {
    int err=write_string(outgoing,request);

    if(err==-1)
      {
       ServerError(client,"Cannot write the outgoing request.");
       PrintMessage(Warning,"Cannot write the outgoing request.");
       exit(1);
      }
   }

 /* Parse the reply */

 if(mode==Real || mode==RealNoCache || mode==Fetch)
   {
    int status=0;

    reply=ParseReply(Url,&status,&replylen);

    if(!reply)
      {
       if(mode==Real || mode==Fetch)
         {
          lseek(spool,0,SEEK_SET);
          ftruncate(spool,0);
          RemoteHostError(spool,Url->name,"the server timed out before sending the reply.");
         }
       if(mode==Real || mode==RealNoCache)
          RemoteHostError(client,Url->name,"the server timed out before sending the reply.");
       PrintMessage(Warning,"Timed out reading the reply.");
       exit(1);
      }

    if(mode==Fetch && (status==301 || status==302))
      {
       char *new_request;

       PrintMessage(Inform,"Server page has been moved.");

       new_request=RequestMoved(Url,request,reply);
       if(!new_request)
          PrintMessage(Warning,"Cannot parse the reply for the new location.");
       else
         {
          int new_outgoing=OpenOutgoingSpoolFile(0);
          if(new_outgoing==-1)
             PrintMessage(Warning,"Cannot open the new outgoing request to write.");
          else
            {
             write_string(new_outgoing,new_request);
             CloseOutgoingSpoolFile(new_outgoing,Url);
             free(new_request);
             fetch_again++;
            }
         }
      }
    else if(status==304)
      {
       PrintMessage(Inform,"Server page is not newer than the one in cache.");
       if(mode==Fetch)
          exit(0);
       if(mode==Real)
          mode=Spool;
      }

    if(replylen<=256)
       reply=(char*)realloc(reply,257);
   }
 else
    reply=(char*)malloc(257);

 /* Close the outgoing file if any. */

 if(outgoing>=0)
   {
    if(mode==Fetch)
       close(outgoing);
    if(mode==SpoolGet || mode==SpoolRefresh || mode==SpoolPragma)
       CloseOutgoingSpoolFile(outgoing,Url);
   }

 /* Read reply and process it. */

 if(mode==Real || mode==RealNoCache)
   {
    int n=replylen,err=0;

    do
      {
       if(mode==Real)
          write_data(spool,reply,n);
       err=write_data(client,reply,n);
      }
    while(err!=-1 && (n=(Url->Protocol->read)(reply,256))>0);

    if(mode==Real)
       if(err==-1 || n<0)
         {
          lseek(spool,0,SEEK_SET);
          ftruncate(spool,0);
          if(err==-1)
            {
             RemoteHostError(spool,Url->name,"the connection was closed by the client during the transfer.");
             PrintMessage(Warning,"Error writing to client [%!s]; client disconnected?.");
             exit(1);
            }
          else
            {
             RemoteHostError(spool,Url->name,"the server timed out during the transfer.");
             PrintMessage(Warning,"Timed out while reading from remote host.");
             exit(1);
            }
         }
       else
         {
          int tell=lseek(spool,0,SEEK_CUR);
          if(tell!=-1)
             ftruncate(spool,tell);
         }
   }
 else if(mode==Fetch)
   {
    int n=replylen;

    do
      {
       write_data(spool,reply,n);
      }
    while((n=(Url->Protocol->read)(reply,256))>0);

    if(n<0)
      {
       lseek(spool,0,SEEK_SET);
       ftruncate(spool,0);
       RemoteHostError(spool,Url->name,"the server timed out during the transfer.");
       PrintMessage(Warning,"Timed out while reading from remote host.");
       exit(1);
      }
    else
      {
       int tell=lseek(spool,0,SEEK_CUR);
       if(tell!=-1)
          ftruncate(spool,tell);
      }

    if(FetchImages || FetchFrames)
      {
       lseek(spool,0,SEEK_SET);

       if(ParseHTML(spool,Url,0))
         {
          char **list;

          if(FetchImages && (list=ListImages()))
             for(n=0;list[n];n++)
               {
                char *new_request=RequestURL(list[n],Url->name);
                int new_outgoing=OpenOutgoingSpoolFile(0);

                PrintMessage(Debug,"Image=%s",list[n]);

                if(new_outgoing==-1)
                   PrintMessage(Warning,"Cannot open the new outgoing request to write.");
                else
                  {
                   write_string(new_outgoing,new_request);
                   CloseOutgoingSpoolFile(new_outgoing,Url);
                   fetch_again++;
                  }

                free(new_request);
               }

          if(FetchFrames && (list=ListFrames()))
             for(n=0;list[n];n++)
               {
                char *new_request=RequestURL(list[n],Url->name);
                int new_outgoing=OpenOutgoingSpoolFile(0);

                PrintMessage(Debug,"Frame=%s",list[n]);

                if(new_outgoing==-1)
                   PrintMessage(Warning,"Cannot open the new outgoing request to write.");
                else
                  {
                   write_string(new_outgoing,new_request);
                   CloseOutgoingSpoolFile(new_outgoing,Url);
                   fetch_again++;
                  }

                free(new_request);
               }
         }
      }
   }
 else if(mode==Spool || mode==SpoolPragma)
   {
    struct stat buf;
    char *head="HTTP/1.0 503 WWWOFFLE Remote Host Error\r\n"; /* This line must not be changed (see messages.c). */
    int n=read_data(spool,reply,64);

    if(!strncmp(reply,head,strlen(head)) || (!fstat(spool,&buf) && buf.st_size==0))
       DeleteWebpageSpoolFile(Url,0);

    if(AddInfoRefresh)
      {
       struct stat buf;
       int status=0;

       sscanf(reply,"%*s %d",&status);

       lseek(spool,0,SEEK_SET);

       if((status>=200 && status<400) && ParseHTML(spool,Url,0) && !fstat(spool,&buf))
         {
          time_t t_ago=time(NULL)-buf.st_mtime;
          int position=GetHTMLEnd();
          char *date=RFC822Date(buf.st_mtime,0),ago[24],*localhost=GetLocalHost(1);
          char *refresh=(char*)malloc(3*strlen(Url->name)+3*strlen(localhost)+192);
          char *index=(char*)malloc(strlen(Url->name)+strlen(localhost)+64);
          char *msg=(char*)malloc(strlen(date)+4*strlen(Url->name)+4*strlen(localhost)+256);

          if(t_ago<0)
             sprintf(ago,"in the future");
          else if(t_ago<3600)
             sprintf(ago,"%ld min%s ago",t_ago/60,t_ago/60==1?"":"s");
          else if(t_ago<(24*3600))
             sprintf(ago,"%ld hour%s ago",t_ago/3600,t_ago/3600==1?"":"s");
          else if(t_ago<(14*24*3600))
             sprintf(ago,"%ld day%s ago",t_ago/(24*3600),t_ago/(24*3600)==1?"":"s");
          else if(t_ago<(30*24*3600))
             sprintf(ago,"%ld week%s ago",t_ago/(7*24*3600),t_ago/(7*24*3600)==1?"":"s");
          else
             sprintf(ago,"%ld month%s ago",t_ago/(30*24*3600),t_ago/(30*24*3600)==1?"":"s");

          if(Url->args && *Url->args=='!')
             sprintf(refresh,"[<a href=\"http://%s/control/delete?url=%s\">Delete</a>"
                             "|Refresh|"
                             "Options]",
                     localhost,Url->name);
          else
             sprintf(refresh,"[<a href=\"http://%s/control/delete?url=%s\">Delete</a>"
                             "|<a href=\"http://%s/refresh/%s/%s\">Refresh</a>|"
                             "<a href=\"http://%s/refresh/?%s\">Options</a>]",
                     localhost,Url->name,
                     localhost,Url->proto,Url->hostp,
                     localhost,Url->name);

          sprintf(index,"<a href=\"http://%s/index/%s/%s/?none\">[Index]</a>",
                  localhost,Url->proto,Url->host);

          sprintf(msg,"\n"
                      "<hr>\n"
                      "<p align=center>\n"
                      "WWWOFFLE - %s (%s) - %s - %s - WWWOFFLE\n"
                      "</p>\n"
                      "<hr>\n",
                  date,ago,refresh,index);

          lseek(spool,0,SEEK_SET);
          while((reply=read_line_or_timeout(spool,reply)))
            {
             if(!strncmp(reply,"Content-Length:",15))
                sprintf(reply,"Content-Length: %d\r\n",atoi(&reply[16])+strlen(msg));
             write_string(client,reply);
             if(reply[0]=='\r' || reply[0]=='\n')
                break;
            }
          reply=realloc(reply,257);

          while(position>0 && (n=read_data(spool,reply,position>256?256:position))>0)
            {
             write_data(client,reply,n);
             position-=n;
            }

          write_string(client,msg);

          while((n=read_data(spool,reply,256))>0)
             write_data(client,reply,n);

          free(localhost);
          free(refresh);
          free(index);
          free(msg);
         }
       else
         {
          lseek(spool,0,SEEK_SET);
          while((n=read_data(spool,reply,256))>0)
             write_data(client,reply,n);
         }
      }
    else
       do
         {
          write_data(client,reply,n);
         }
       while((n=read_data(spool,reply,256))>0);
   }
 else if(mode==SpoolGet)
   {
    WillGetURL(client,Url->name,0);
   }
 else if(mode==SpoolWillGet)
   {
    WillGetURL(client,Url->name,1);
   }
 else if(mode==RealRefresh || mode==SpoolRefresh)
   {
    RefreshRedirect(client,Url->name);
   }

 /* Close down and exit. */

 if(mode==Real && ExistsOutgoingSpoolFile(Url))
    DeleteOutgoingSpoolFile(Url);

 if(mode==Real || mode==RealNoCache || mode==Fetch)
    (Url->Protocol->close)();

 if(spool>=0)
    close(spool);

 if(request)
    free(request);

 if(reply)
    free(reply);

 if(client>=0)
    CloseSocket(client);

 if(fetch_again)
    return(4);
 else
    return(0);
}


/*++++++++++++++++++++++++++++++++++++++
  Uninstall the signal handlers.
  ++++++++++++++++++++++++++++++++++++++*/

static void uninstall_sighandlers(void)
{
 struct sigaction action;

 /* SIGCHLD */
 action.sa_handler = SIG_DFL;
 sigemptyset (&action.sa_mask);
 action.sa_flags = 0;
 if(sigaction(SIGCHLD, &action, NULL) != 0)
    PrintMessage(Warning, "Cannot uninstall SIGCHLD handler.");

 /* SIGINT, SIGQUIT, SIGTERM */
 action.sa_handler = SIG_DFL;
 sigemptyset(&action.sa_mask);
 action.sa_flags = 0;
 if(sigaction(SIGINT, &action, NULL) != 0)
    PrintMessage(Warning, "Cannot uninstall SIGINT handler.");
 if(sigaction(SIGQUIT, &action, NULL) != 0)
    PrintMessage(Warning, "Cannot uninstall SIGQUIT handler.");
 if(sigaction(SIGTERM, &action, NULL) != 0)
    PrintMessage(Warning, "Cannot uninstall SIGTERM handler.");

 /* SIGHUP */
 action.sa_handler = SIG_DFL;
 sigemptyset(&action.sa_mask);
 action.sa_flags = 0;
 if(sigaction(SIGHUP, &action, NULL) != 0)
    PrintMessage(Warning, "Cannot uninstall SIGHUP handler.");

 /* SIGPIPE */
 action.sa_handler = SIG_IGN;
 sigemptyset (&action.sa_mask);
 action.sa_flags = 0;
 if(sigaction(SIGPIPE, &action, NULL) != 0)
    PrintMessage(Warning, "Cannot ignore SIGPIPE.");
}
